/*============================================================================
  File:     Estimates.sql

  Summary:  Where do estimates come from?
				Existing statistics?
				Auto-created statistics?
				Heuristics?
  
  SQL Server Version: SQL Server 2008+
------------------------------------------------------------------------------
  Written by Kimberly L. Tripp, SQLskills.com

  For more scripts and sample code, check out 
    http://www.SQLskills.com

  This script is intended only as a supplement to demos and lectures
  given by Kimberly L. Tripp.  
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

-- These samples use the Credit database. You can download and restore the
-- credit database from here:
-- http://www.sqlskills.com/sql-server-resources/sql-server-demos/ 

-- NOTE: You can use a SQL Server 2000 backup and restore to: 2000, 2005 or 2008/R2
-- There's also a 2008 backup that you can restore to 2008/R2 or 2012

---------------------------------------------
-- Even distribution
---------------------------------------------

USE [Credit];
GO

-- Quick/general way to see index list
EXEC [sp_helpindex] 'dbo.member';
GO

-- Rewrite to give better index details/internals
-- Get here: http://www.sqlskills.com/blogs/kimberly/use-this-new-sql-server-2012-rewrite-for-sp_helpindex/ 
EXEC [sp_SQLskills_SQL2012_helpindex] 'dbo.member';
GO

EXEC [sp_helpstats] 'dbo.member', 'all';
GO

CREATE INDEX [membername]
ON [dbo].[member]
([lastname], [firstName], [middleinitial]);
GO

SET STATISTICS IO ON;
-- Turn on showplan 
-- Use: Query, Include Actual Execution Plan [Ctrl + M]
GO

SELECT [m].* 
FROM [dbo].[member] AS [m]
WHERE [m].[lastname] = 'Anderson';  
GO

-- Where did this come from?
--	Key: adhoc, can be sniffed, uses the histogram
--	Value is an actual step - uses EQ_ROWS
DBCC SHOW_STATISTICS('dbo.member', 'membername');
GO

-- What about a value that's NOT in as a step:
SELECT [m].* 
FROM [dbo].[member] AS [m]
WHERE [m].[lastname] = 'Smith';  
GO

-- Where did this come from?
--	Key: adhoc, can be sniffed, uses the histogram
--	Value is in the range: RYAN - STEIN
--  uses AVG_RANGE_ROWS
DBCC SHOW_STATISTICS('dbo.member', 'membername')
WITH HISTOGRAM;
GO

-- Finally, can we see that this data is EVENLY 
-- distributed:
-- Yes - in the steps of the histogram
--   All of the EQ_ROWS are basically the same
--   All of the RANGE_ROWS are 0 (no rows between
--       step values
-- Finally, what about the density vector:
-- Take All density * rows = average rows returned

SELECT 0.03846154	-- All density: for lastname
	    * 10000		-- Rows: from the header
--  = 384.61540000	-- On average, you will get back
					-- 385 rows for a single lastname
					-- supplied in a query

---------------------------------------------
-- Change some data
---------------------------------------------

UPDATE [dbo].[member] 
	SET [lastname] = 'Tripp'
WHERE [member_no] = 1234;  
GO

SELECT [m].* 
FROM [dbo].[member] AS [m]
WHERE [m].[lastname] = 'Tripp';  
GO

-- Where did this come from?
--	Key: adhoc, can be sniffed, uses the histogram
--	Value is in the range: TOLSKY - UNAKLET
-- Obviously, NOT in a step
-- But, OK, because not much has changed and
-- this still works in AVG_RANGE_ROWS
DBCC SHOW_STATISTICS('dbo.member', 'membername')
WITH HISTOGRAM;
GO

-- Update another 1000 rows:
UPDATE [dbo].[member] 
	SET [lastname] = 'Tripp'
WHERE [member_no] > 9000 ;  
GO

SET STATISTICS TIME ON;
GO

SELECT [m].* 
FROM [dbo].[member] AS [m]
WHERE [m].[lastname] = 'Tripp';  
-- Estimate didn't change
GO

-- Compare this to a table scan
SELECT [m].* 
FROM [dbo].[member] AS [m] WITH (INDEX (0))
WHERE [m].[lastname] = 'Tripp';  
-- Estimate didn't change


-- Tangent on when stats change:
-- Read the whitepaper: http://msdn.microsoft.com/en-us/library/dd535534.aspx

-- Quick summary of When they change:
-- DEFAULT: 500 + 20% of the table (member = 10K rows)
-- Member's stats will get INVALIDATED at 2,500 rows
--						20% = 2000 + 500

-- Or, if trace flag 2371 is turned on, then
-- updates are dynamic (tied to table size)
-- http://blogs.msdn.com/b/saponsqlserver/archive/2011/09/07/changes-to-automatic-update-statistics-in-sql-server-traceflag-2371.aspx

---------------------------------------------
-- Force stats to update:
---------------------------------------------

UPDATE STATISTICS [dbo].[member];
-- This updates all statistics for the table
-- Fullscan for scans under 10K pages
-- Sampling for scans at 10K or over 
-- Not fixed sampling percentage: percentage 
--     varies as scan size increases
GO

SELECT [m].* 
FROM [dbo].[member] AS [m]
WHERE [m].[lastname] = 'Tripp';  
-- Notice my estimate changed? Why?
GO

-- Plenty of room for more steps (up to 200)
-- And, SQL Server saw Tripp as a new/interesting
-- value:
DBCC SHOW_STATISTICS('dbo.member', 'membername')
WITH HISTOGRAM;
GO

---------------------------------------------
-- What if I use a variable?
---------------------------------------------

DECLARE @LastNameVar	varchar(15) = 'Tripp'
		-- this is the EXACT same data type as the column
SELECT [m].* 
FROM [dbo].[member] AS [m]
WHERE [m].[lastname] = @LastNameVar;
GO

-- What? 
--  The estimate went to: 370.37
-- Why?
--  During compilation the variable was deemed "UNKNOWN"
--	(which makes sense, the assignment doesn't happen
--  until runtime/execution)

-- If a value is unknown - it cannot be "sniffed"
-- If it cannot be sniffed then SQL Server has to
-- use the "average" number of rows. This comes
-- from the density vector.

-- Take All density * rows = average rows returned

SELECT 0.03703704	-- All density: for lastname
	    * 10000		-- Rows: from the header
--  = 370.37040000	-- On average, you will get back
					-- 370.37 rows for a single last name
					-- supplied in a query
DBCC SHOW_STATISTICS('dbo.member', 'membername');
GO

-- But, now we're starting to show some "skew"
-- in the data. And, it's starting to become
-- a bit more obvious that when a value CAN
-- be sniffed then for THAT value and THAT
-- query - you'll usually get a better plan
-- (for THAT execution)

---------------------------------------------
-- What if we don't have statistics?
---------------------------------------------

-- First, let's the drop index we've been using
DROP INDEX [dbo].[member].[membername];
GO

-- Then, let's turn off auto create statistics
-- Generally speaking - this is NOT recommended! 
ALTER DATABASE [Credit] 
SET AUTO_CREATE_STATISTICS OFF
go

-- Check to see if there are ANY stats that could
-- be used for lastname:
EXEC [sp_SQLskills_SQL2012_helpindex] 'dbo.member';
EXEC [sp_helpstats] 'dbo.member', 'all';
GO

-- No indexes or stats that even have lastname IN
-- the index (not even in INCLUDE).

-- Literal/parameter
SELECT [m].* 
FROM [dbo].[member] AS [m]
WHERE [m].[lastname] = 'Anderson';  
GO

-- Variable
DECLARE @LastNameVar	varchar(15) = 'Tripp'
		-- this is the EXACT same data type as the column
SELECT [m].* 
FROM [dbo].[member] AS [m]
WHERE [m].[lastname] = @LastNameVar;
GO

-- The plans are the same
-- The estimates are the same (10%)

-- No statistics therefore HEURISTICS 
-- Heuristics = internal rules

-- Rule for equality on one column is:
--   # of rows in the table to the power of 3/4
--   10000 ^ 3/4 = 1000

-- The rules for more than 2 columns are fractionally
-- smaller (and more complicated). But, the main point
-- statistics are generally better than heuristics.

-- Other things to note:
--	The warning symbol
--	Columns with no statistics

---------------------------------------------
-- Other Heuristics
---------------------------------------------

-- What about > or < ?
DECLARE @test   int = 9999
SELECT [m].* 
FROM [dbo].[member] AS [m]
WHERE [m].[member_no] > @test;
		-- A variable cannot be sniffed 
		-- With an unknown value, they need to guess? 
		-- It's not a guess, it's heuristics.
GO

DECLARE @test   int = 2
SELECT [m].* 
FROM [dbo].[member] AS [m]
WHERE [m].[member_no] > @test;
		-- Again, heuristics.
go

-- Again - they estimate 30% 
DECLARE @test   int = 1
SELECT [m].* 
FROM [dbo].[member] AS [m]
WHERE [m].[member_no] < @test;
		-- Again, heuristics.
go

-- Here - they yield 9% *no matter what*
DECLARE @lowerbound   int = 1
DECLARE @upperbound   int = 10000

SELECT [m].* 
FROM [dbo].[member] AS [m]
WHERE [m].[member_no] BETWEEN @lowerbound 
						AND @upperbound;
GO

-- What about a comparison between columns?
IF OBJECT_ID('[dbo].[member2]') IS NOT NULL
	DROP TABLE [dbo].[member2];
GO

SELECT [m].[member_no] AS [MNO1]
	, [m].[member_no] AS [MNO2]
	, [m].[lastname]
	, [m].[firstname]
	, [m].[middleinitial]
INTO [dbo].[member2]
FROM [dbo].[member] AS [m];
GO

SELECT [m].* 
FROM [dbo].[member2] AS [m]
WHERE [m].[MNO1] = [m].[MNO2] 
-- Again, heuristics
-- A comparison between columns is 10%

-- The prior examples/comments are for 2008/2008R2 and 2012

-- NOTE: SQL Server 2014 has made changes to the internal
-- heuristic rules. Everything prior to "line 224: What if 
-- we don't have statistics?" is the SAME in 2014
-- but what's below is different (and, yes, better) in 2014.
-- Having said that, even though some of these heuristics
-- might be better for THESE cases below, it doesn't mean
-- that I want you to use heuristics. Statistics are ALWAYS
-- better than heuristics. Do not turn off auto create 
-- statistics.
